From mathcomp Require Import boot finite_group.
Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.
Local Open Scope group_scope.
Open Scope group_scope.
Check @ker_sdprodm.
